# -*- coding: utf-8 -*-
import asyncio
from unittest import mock
from urllib.error import URLError

import click

from pytest import mark

from story import api, cli

from websockets import ConnectionClosed, InvalidStatusCode


@mark.parametrize('level', [None, 'info', 'debug', 'warning', 'error'])
@mark.parametrize('follow', [True, False])
@mark.parametrize('last', [10, 100, 4000, None])
@mark.parametrize('service,service_name', [
    (False, None),
    (True, None),
    (True, 'my_service'),
])
def test_logs_command(runner, init_sample_app_in_cwd, patch, level, follow,
                      last, service, service_name):
    patch.object(asyncio, 'get_event_loop')
    patch.object(cli, 'track')
    patch.object(api.Apps, 'get_uuid_from_hostname')

    args = []

    if service:
        args.append('-s')

    if service_name:
        args.append('-sn')
        args.append(service_name)

    if follow:
        args.append('-f')

    if level:
        args.append('-l')
        args.append(level)

    if last:
        args.append('-n')
        args.append(last)

    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()

        from story.commands import logs
        patch.object(logs, 'connect_and_listen_with_retry')

        result = runner.run(logs.logs, args=args)

    assert 'Retrieving logs for my_app' in result.stdout

    api.Apps.get_uuid_from_hostname.assert_called_with('my_app')

    expected_service_name = 'N/A'
    if service:
        if service_name:
            expected_service_name = service_name
        else:
            expected_service_name = '*'

    cli.track.assert_called_with('App Logs Requested', {
        'App name': 'my_app',
        'Follow': 'Yes' if follow else 'No',
        'Last N': last if last else 10,
        'Source': 'Service' if service else 'Runtime',
        'Service Name': expected_service_name,
        'Minimum level': 'Info' if level is None else level.capitalize()
    })

    logs.connect_and_listen_with_retry.assert_called_with(
        api.Apps.get_uuid_from_hostname.return_value,
        last if last else 10,
        follow,
        service is False,
        service,
        '*' if service_name is None else service_name,
        'info' if level is None else level
    )

    asyncio.get_event_loop.return_value.run_until_complete.assert_called_with(
        logs.connect_and_listen_with_retry.return_value
    )


@mark.parametrize('exception_to_throw', [URLError('reason'),
                                         ConnectionClosed(10, 'closed')])
@mark.asyncio
async def test_ping_forever(runner, init_sample_app_in_cwd, magic, patch,
                            async_mock, exception_to_throw):
    websocket = magic()
    patch.object(asyncio, 'sleep', new=async_mock())
    patch.object(websocket, 'send', new=async_mock(side_effect=[
        None, None, None, None, exception_to_throw]))
    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()
        from story.commands.logs import ping_forever
        await ping_forever(websocket)

    assert websocket.send.mock.mock_calls == [
        mock.call('{"command": "ping"}'),
        mock.call('{"command": "ping"}'),
        mock.call('{"command": "ping"}'),
        mock.call('{"command": "ping"}'),
        mock.call('{"command": "ping"}')
    ]

    assert asyncio.sleep.mock.call_count == 4


@mark.parametrize('final_result_of_connect_once', [
    URLError('reason'),
    InvalidStatusCode(10),
    InvalidStatusCode(500),
    True
])
@mark.asyncio
async def test_connect_and_listen_with_retry(runner, init_sample_app_in_cwd,
                                             patch, async_mock,
                                             final_result_of_connect_once):
    patch.object(click, 'get_current_context')

    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()

        from story.commands import logs

        patch.object(logs, 'connect_and_listen_once',
                     new=async_mock(side_effect=[
                         False,
                         False,
                         False,
                         False,
                         final_result_of_connect_once]))

        await logs.connect_and_listen_with_retry(
            'app_id', 'n', 'follow', 'runtime_logs', 'service_logs',
            'service_name', 'level')

    assert logs.connect_and_listen_once.mock.mock_calls == [
        mock.call('app_id', 'n', 'follow', 'runtime_logs', 'service_logs',
                  'service_name', 'level'),
        mock.call('app_id', 100, 'follow', 'runtime_logs', 'service_logs',
                  'service_name', 'level'),
        mock.call('app_id', 100, 'follow', 'runtime_logs', 'service_logs',
                  'service_name', 'level'),
        mock.call('app_id', 100, 'follow', 'runtime_logs', 'service_logs',
                  'service_name', 'level'),
        mock.call('app_id', 100, 'follow', 'runtime_logs', 'service_logs',
                  'service_name', 'level')
    ]

    if isinstance(final_result_of_connect_once, BaseException):
        click.get_current_context.return_value.exit.assert_called_with(1)



# # -*- coding: utf-8 -*-
#
# from json import dumps
#
#
# class Scope:
#     def __init__(self):
#         self._levels = [{}]
#
#     def __len__(self):
#         return len(self._levels)
#
#     def __contains__(self, key):
#         for c in self._levels:
#             if key in c:
#                 return True
#         return False
#
#     def pop(self):
#         self._levels.pop(0)
#
#     def add(self):
#         self._levels.insert(0, {})
#
#     def update(self, data):
#         self._levels[-1].update(data)
#
#     def dumps(self):
#         # TODO merge a list of keys
#         return dumps(self._levels, indent=4)
#
#     def __getitem__(self, key):
#         for c in self._levels:
#             if key in c:
#                 return c[key]
#         raise KeyError(f'UndefinedVariable "{key}"')
#
#     def indent(self):
#         return ' ' * (4 * (len(self) - 1))



# -*- coding: utf-8 -*-
import os

import click

import click_completion

from .. import cli


def custom_startswith(string, incomplete):
    """A custom completion matching that supports case insensitive matching"""
    if os.environ.get('_CLICK_COMPLETION_COMMAND_CASE_INSENSITIVE_COMPLETE'):
        string = string.lower()
        incomplete = incomplete.lower()
    return string.startswith(incomplete)


click_completion.core.startswith = custom_startswith
click_completion.init()


@cli.cli.command()
@click.option(
    '-i',
    '--case-insensitive',
    is_flag=True,
    help='Case insensitive completion',
)
@click.option(
    '--install',
    is_flag=True,
    help='Install completion code into shell startup file',
)
@click.argument(
    'shell',
    required=False,
    type=click_completion.DocumentedChoice(click_completion.core.shells),
)
@click.option('--path', required=False,
              help='Specify a custom completion file for your shell')
def completion(path, shell, install, case_insensitive):
    """Show or install shell completion code"""
    extra_env = {}

    if case_insensitive:
        extra_env['_CLICK_COMPLETION_COMMAND_CASE_INSENSITIVE_COMPLETE'] = 'ON'

    if install:
        shell, path = click_completion.core.install(
            shell=shell, path=path, append=True, extra_env=extra_env
        )
        click.echo('%s completion installed in %s' % (shell, path))
    else:
        click.echo(click_completion.core.get_code(shell, extra_env=extra_env))



# -*- coding: utf-8 -*-
import sys

from blindspin import spinner

import click

import emoji

from .. import api
from .. import cli
from .. import options


@cli.cli.group()
def config():
    """Update the configuration for your app."""
    pass


@config.command(name='list')
@options.app()
def list_command(app):
    """List environment variables."""

    cli.user()
    click.echo('Fetching config… ', nl=False)

    with spinner():
        config = api.Config.get(app)

    click.echo(
        click.style('\b' + emoji.emojize(':heavy_check_mark:'), fg='green')
    )

    if config:
        click.echo(click.style('Storyscript variables:', dim=True))
        for name, value in config.items():
            if not isinstance(value, dict):
                click.echo(click.style(name, fg='green') + f':  {value}')

        click.echo('')
        click.echo(click.style('Service variables:', dim=True))
        for name, value in config.items():
            if isinstance(value, dict):
                click.echo(click.style(name, bold=True))
                for _name, _value in value.items():
                    click.echo(
                        '  ' + click.style(_name, fg='green') + f':  {_value}'
                    )

    else:
        click.echo(click.style('No configuration set yet.', bold=True))
        click.echo()
        click.echo('Set Storyscript secrets with:')
        cli.print_command('story config set key=value')
        click.echo('Set service environment variables with:')
        cli.print_command('story config set service.key=value')


@config.command(name='set')
@click.argument('variables', nargs=-1, required=True)
@click.option(
    '--message',
    '-m',
    nargs=1,
    default=None,
    help='(optional) Message why variable(s) were created.',
)
@options.app()
def set_command(variables, app, message):
    """
    Set one or more environment variables.

        $ story config set key=value foo=bar

    To set an environment variable for a specific service use

        $ story config set twitter.oauth_token=value

    """
    cli.user()

    click.echo('Fetching config… ', nl=False)
    with spinner():
        config = api.Config.get(app=app)
    click.echo(
        click.style('\b' + emoji.emojize(':heavy_check_mark:'), fg='green')
    )

    for keyval in variables:
        try:
            key, val = tuple(keyval.split('=', 1))
        except ValueError:
            click.echo(
                f'Config variables must be of the form name=value.'
                f'\nGot unexpected pair "{keyval}"',
                err=True,
            )
            click.echo(set_command.__doc__.strip())
            sys.exit(1)

        if '.' in key:
            service, key = tuple(key.split('.', 1))
            config.setdefault(service.lower(), {})[key.upper()] = val
        else:
            config[key.upper()] = val

        click.echo()
        click.echo(f" {click.style(key.upper(), fg='green')}: {val}")

    click.echo('\nSetting config and deploying new release…  ', nl=False)
    with spinner():
        release = api.Config.set(config=config, app=app, message=message)
    click.echo(
        click.style('\b' + emoji.emojize(':heavy_check_mark:'), fg='green')
    )
    click.echo(
        f'Deployed new release… '
        + click.style(f'v{release["id"]}', bold=True, fg='magenta')
    )


@config.command()
@click.argument('variables', nargs=-1, required=True)
@options.app()
def get(variables, app):
    """Get one or more environment variables."""

    cli.user()
    click.echo(f'Fetching config for {app}… ', nl=False)

    with spinner():
        config = api.Config.get(app=app)

    click.echo(
        click.style(emoji.emojize(':heavy_check_mark:'), fg='green')
    )

    for name in variables:
        if '.' in name:
            service, name = tuple(name.split('.', 1))
            value = config.get(service.lower(), {}).get(name.upper(), None)
        else:
            if name in config:
                # could be a service here
                value = config[name]
            else:
                value = config.get(name.upper(), None)

        if value:
            if isinstance(value, dict):
                for name, value in value.items():
                    click.echo(
                        click.style(name.upper(), fg='green')
                        + f':  {value}'
                    )
            else:
                click.echo(
                    click.style(name.upper(), fg='green') + f':  {value}'
                )
        else:
            click.echo(
                click.style(
                    f'No variable named "{name.upper()}".', fg='red'
                )
            )


@config.command(name='del')
@click.argument('variables', nargs=-1, required=True)
@click.option(
    '--message',
    '-m',
    nargs=1,
    default=None,
    help='(optional) Message why variable(s) were deleted.',
)
@options.app()
def del_command(variables, app, message):
    """
    Delete one or more environment variables
    """
    cli.user()
    click.echo('Fetching config… ', nl=False)
    with spinner():
        config = api.Config.get(app=app)
    click.echo(
        click.style(emoji.emojize(':heavy_check_mark:'), fg='green')
    )

    for key in variables:
        removed = False
        if key in config:
            if type(config.pop(key)) is dict:
                click.echo(
                    click.style('Removed service', fg='red') + f': {key}'
                )
            else:
                removed = True
        elif key.upper() in config:
            config.pop(key.upper())
            removed = True
        elif '.' in key:
            service, key = tuple(key.split('.', 1))
            if service in config and key.upper() in config[service]:
                config[service].pop(key.upper())
                removed = True

        if removed:
            click.echo(
                click.style('Removed', fg='red') + f': {key.upper()}'
            )

    click.echo('\nSetting config and deploying new release… ', nl=False)
    with spinner():
        release = api.Config.set(config=config, app=app, message=message)
    click.echo(
        click.style('\b' + emoji.emojize(':heavy_check_mark:'), fg='green')
    )
    click.echo(
        f'Deployed new release… '
        + click.style(f'v{release["id"]}', bold=True, fg='magenta')
    )



# -*- coding: utf-8 -*-
import sys

from blindspin import spinner

import click

import emoji

from .. import api
from .. import cli
from .. import options
from ..helpers import datetime


@cli.cli.group()
def releases():
    """Manage releases for your app (including rollback)."""
    pass


@releases.command(name='list')
@click.option(
    '--limit', '-n', nargs=1, default=20, help='List N latest releases'
)
@options.app()
def list_command(app, limit):
    """List application releases."""
    cli.user()

    # click.echo(click.style('Releases', fg='magenta'))
    # click.echo(click.style('========', fg='magenta'))

    with spinner():
        res = api.Releases.list(app, limit=limit)

    res = sorted(res, key=lambda elem: elem['id'])

    if res:
        from texttable import Texttable

        table = Texttable(max_width=800)
        table.set_deco(Texttable.HEADER)
        table.set_cols_align(['l', 'l', 'l', 'l'])
        all_releases = [['VERSION', 'STATUS', 'CREATED', 'MESSAGE']]
        for release in res:
            date = datetime.parse_psql_date_str(release['timestamp'])
            all_releases.append(
                [
                    f'v{release["id"]}',
                    release['state'].capitalize(),
                    datetime.reltime(date),
                    release['message'],
                ]
            )
        table.add_rows(rows=all_releases)
        click.echo(table.draw())
    else:
        click.echo(f'No releases yet for app {app}.')


@releases.command()
@click.argument('version', nargs=1, required=False)
@options.app()
def rollback(version, app):
    """Rollback release to a previous release."""
    cli.user()

    if version and version[0] == 'v':
        version = version[1:]

    if not version:
        click.echo(f'Getting latest release for app {app}…  ', nl=False)
        with spinner():
            res = api.Releases.get(app=app)
            version = int(res[0]['id']) - 1
        click.echo(
            click.style('\b' + emoji.emojize(':heavy_check_mark:'), fg='green')
        )

    try:
        if int(version) <= 0:
            click.echo('Unable to rollback a release before v1.')
            sys.exit(1)
    except ValueError:
        click.echo(
            click.style('Invalid release specified.', fg='red'), err=True
        )
        sys.exit(1)

    click.echo(f'Rolling back to v{version}…  ', nl=False)

    with spinner():
        res = api.Releases.rollback(version=version, app=app)

    click.echo(
        click.style('\b' + emoji.emojize(':heavy_check_mark:'), fg='green')
    )
    click.echo(
        f'Deployed new release… '
        + click.style(f'v{res["id"]}', bold=True, fg='magenta')
    )



# -*- coding: utf-8 -*-
import asyncio
import json
import socket
import sys
import time
from urllib.error import URLError

import click

import click_spinner

import websockets
from websockets import WebSocketClientProtocol

from .. import cli
from .. import options
from ..api import Apps


@cli.cli.command()
@click.option('--follow', '-f', is_flag=True, help='Follow the logs')
@click.option('--last', '-n', default=10, help='Print the last n lines')
@click.option(
    '--services',
    '-s',
    'service',
    is_flag=True,
    help='Return logs from services used in the story',
)
@click.option(
    '--service-name',
    '-sn',
    default='*',
    help='Return logs for a specific service given by '
    'its name (eg: "redis" or "owner/name")',
)
@click.option(
    '--level',
    '-l',
    default='info',
    type=click.Choice(['debug', 'info', 'warning', 'error']),
    help='Specify the minimum log level '
    '(does not work when --services/-s is specified)',
)
@options.app()
def logs(follow, last, service, service_name, app, level):
    """
    Fetch logs for your app
    """
    cli.user()

    click.echo(f'Retrieving logs for {app}... ', nl=False)
    with click_spinner.spinner():
        app_id = Apps.get_uuid_from_hostname(app)

    click.echo()

    cli.track(
        'App Logs Requested',
        {
            'App name': app,
            'Follow': 'Yes' if follow else 'No',
            'Last N': last,
            'Source': 'Runtime' if service is False else 'Service',
            'Service Name': service_name if service else 'N/A',
            'Minimum level': level.capitalize(),
        },
    )

    asyncio.get_event_loop().run_until_complete(
        connect_and_listen_with_retry(
            app_id,
            last,
            follow,
            service is False,
            service,
            service_name,
            level,
        )
    )


async def ping_forever(websocket: WebSocketClientProtocol):
    while True:
        try:
            await websocket.send('{"command": "ping"}')
        except websockets.exceptions.ConnectionClosed:
            return  # Stop looping when this connection is closed.
        except URLError:
            return
        await asyncio.sleep(10)


async def connect_and_listen_with_retry(
    app_id, n, follow, runtime_logs, service_logs, service_name, level
):
    """
    Every 4-5 minutes, the connection terminates. This is
    not intentional, as the upstream server keeps it alive.
    However, something in the middle causes the connection to
    break. So, whenever it breaks, we just reconnect,
    but set N to be 0, since we want to continue from where we
    left off.
    If somebody does ever figure out why it breaks,
    please ping @judepereira on GitHub.
    """
    while True:
        try:
            completed = await connect_and_listen_once(
                app_id,
                n,
                follow,
                runtime_logs,
                service_logs,
                service_name,
                level,
            )
        except (URLError, socket.gaierror):
            click.echo('Network connection lost', err=True)
            click.get_current_context().exit(1)
            return
        except websockets.exceptions.InvalidStatusCode as e:
            if int(e.status_code / 100) == 5:
                click.echo(
                    'The upstream log server appears to be restarting.'
                    '\nPlease try again in a few seconds.',
                    err=True,
                )
            else:
                click.echo(
                    'The upstream log server did not respond.'
                    '\nPlease try again in a few seconds.',
                    err=True,
                )
            click.get_current_context().exit(1)
            return

        if completed:
            break

        # Don't worry, cut_off_ts will take care of
        # not printing duplicate messages.
        # We set n=100 just so that messages are not lost during a
        # connection termination.
        n = 100


cut_off_ts = 0
"""
Logic around cut_off_ts: This field holds the last last log's timestamp,
and is used to ensure that messages displayed by the CLI are not missed
in the event of an abrupt connection termination. See the pydoc for
connect_and_listen_with_retry above for more help.
"""


async def connect_and_listen_once(
    app_id, n, follow, runtime_logs, service_logs, service_name, level
):
    global cut_off_ts

    async with websockets.connect('wss://logs.storyscript.io') as websocket:
        assert isinstance(websocket, WebSocketClientProtocol)

        auth_payload = {
            'command': 'auth',
            'access_token': cli.get_access_token(),
            'id': cli.get_user_id(),
            'app_id': app_id,
        }
        await websocket.send(json.dumps(auth_payload))

        try:
            auth_response = await websocket.recv()

            # Generally, the log server will close the connection instantly.
            # This authorised check is just for consistency.
            # See https://github.com/storyscript/logstreamer#authentication.
            auth_response = json.loads(auth_response)
            if not auth_response['authorised']:
                raise websockets.exceptions.ConnectionClosed(-1, None)

        except websockets.exceptions.ConnectionClosed:
            click.echo(
                'The log server sent an unauthorised response.\n'
                'Are you sure you have access to view '
                'this app\'s logs?',
                err=True,
            )
            sys.exit(1)

        # Keep the connection alive by sending pings.
        asyncio.get_event_loop().create_task(ping_forever(websocket))

        # Send our filter payload.
        if runtime_logs:
            filter_payload = {
                'command': 'filter',
                'source': 'runtime',
                'level': level,
                'n': n,
                'watch': follow,
            }
        else:
            assert service_logs is True
            filter_payload = {
                'command': 'filter',
                'source': 'service',
                'service_name': service_name,
                'level': level,  # Note: The upstream server ignores level.
                'n': n,
                'watch': follow,
            }

        await websocket.send(json.dumps(filter_payload))

        last_log = None
        while True:
            try:
                message = await websocket.recv()
            except websockets.exceptions.ConnectionClosed:
                # Connection was closed since all the required log items were
                # sent.
                if not follow:
                    return True

                cut_off_ts = last_log['ts']
                return False

            log = json.loads(message)
            if log.get('type') == 'error':
                click.echo(
                    message=log.get('message', 'An unknown error occurred'),
                    err=True)
                return True

            last_log = log

            if cut_off_ts >= log['ts']:
                continue

            date = time.strftime(
                '%b %d %H:%M:%S', time.localtime(int(log['ts'] / 1000))
            )

            tag = 'runtime'
            if service_logs:
                tag = log['service_name']

            colourize_and_print(date, tag, log['level'], log['message'])

    return True


def colourize_and_print(date, tag, level, message):
    level_col = 'green'  # Default for info.
    level = level.lower()
    if 'debug' in level:
        level_col = 'blue'
    elif 'warn' in level:
        level_col = 'yellow'
    elif 'crit' in level or 'error' in level:
        level_col = 'red'

    level = level[:7].rjust(7)
    tag = tag[:12].rjust(12)

    if tag:
        click.echo(
            f'{click.style(date, fg="white")} '
            f'{click.style(level.upper(), fg=level_col)} '
            f'{click.style(tag, fg="blue")}: '
            f'{message}'
        )
    else:
        click.echo(
            f'{click.style(date, fg="white")} '
            f'{click.style(level.upper(), fg=level_col)} '
            f'{message}'
        )



# -*- coding: utf-8 -*-
from pkg_resources import DistributionNotFound, get_distribution

try:
    version = get_distribution('story').version
    compiler_version = get_distribution('storyscript').version
except DistributionNotFound:
    # package is not installed
    version = 'unknown'
    compiler_version = 'unknown'



# -*- coding: utf-8 -*-
from pytest import mark


@mark.parametrize('enabled', [True, False])
def test_check(runner, patch, enabled, init_sample_app_in_cwd):
    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()
        from story import api
        from story.commands import maintenance

        patch.object(api.Apps, 'maintenance', return_value=enabled)

        result = runner.run(maintenance.check, exit_code=0)

    assert 'Fetching maintenance mode' in result.stdout
    if enabled:
        assert 'ON. Application is disabled.' in result.stdout
    else:
        assert 'off. Application is running.' in result.stdout

    api.Apps.maintenance.assert_called_with(app='my_app', maintenance=None)


@mark.parametrize('enable,assert_messages', [
    (True, ['Enabling maintenance mode for app my_app',
            'Application is now in maintenance–mode']),
    (False, ['Disabling maintenance mode for app my_app',
             'Application is now running.']),
])
def test_on_and_off(runner, patch, init_sample_app_in_cwd, enable,
                    assert_messages):
    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()
        from story import api
        from story.commands import maintenance

        patch.object(api.Apps, 'maintenance')

        if enable:
            result = runner.run(maintenance.on, exit_code=0)
        else:
            result = runner.run(maintenance.off, exit_code=0)

    for m in assert_messages:
        assert m in result.stdout

    api.Apps.maintenance.assert_called_with(app='my_app', maintenance=enable)



# -*- coding: utf-8 -*-
import json
import os
import sys

import click

import emoji

from storyscript.exceptions import StoryError

from .. import cli, options, utils


@cli.cli.command()
@click.option('--debug', is_flag=True, help='Compile in debug mode')
@options.app()
def test(debug, app):
    """Compile your Storyscripts, and check for any errors."""

    cli.user()

    app_name = cli.get_app_name_from_yml() or 'Not created'
    tree = compile_app(app_name, debug)

    if tree is None:
        sys.exit(1)

    count = len(tree.get('stories', {}))

    if count == 0:
        click.echo(click.style('\tX', fg='red') + ' No stories found')
        sys.exit(1)
    else:
        for k, v in tree['stories'].items():
            click.echo(
                ' - '
                + f'{k}'
                + click.style(
                    f'{emoji.emojize(" :heavy_check_mark:")}', fg='green'
                )
            )

    click.echo()
    click.echo(emoji.emojize('Looking good! :party_popper:'))
    click.echo()
    click.echo('Deploy your app with:')
    cli.print_command('story deploy')


def compile_app(app_name_for_analytics, debug) -> dict:
    """Compiles, prints pretty info, and returns the compiled tree.
    :return: The compiled tree
    """
    from storyscript.App import App

    click.echo(click.style('Compiling Stories… ', bold=True))

    old_cwd = os.getcwd()

    try:
        os.chdir(utils.get_project_root_dir())
        stories = json.loads(App.compile(utils.get_project_root_dir()))
    except StoryError as e:
        click.echo('Failed to compile project:\n', err=True)
        click.echo(click.style(str(e.message()), fg='red'), err=True)
        stories = None
    except BaseException as e:
        click.echo('Failed to compile project:\n', err=True)
        click.echo(click.style(str(e), fg='red'), err=True)
        stories = None
    finally:
        os.chdir(old_cwd)

    result = 'Success'
    count = 0

    if stories is None:
        result = 'Failed'
    else:
        count = len(stories.get('stories', {}))
        stories['yaml'] = cli.get_asyncy_yaml()

    cli.track(
        'App Compiled',
        {
            'App name': app_name_for_analytics,
            'Result': result,
            'Stories': count,
        },
    )

    return stories



import os


def find_story_yml():
    """Finds './story.yml'."""
    path = _find_story_yml('story.yml')
    if not path:
        path = _find_story_yml('asyncy.yml')

    return path


def _find_story_yml(file_name):
    current_dir = os.getcwd()
    while True:
        if os.path.exists(f'{current_dir}{os.path.sep}{file_name}'):
            return f'{current_dir}{os.path.sep}{file_name}'
        elif current_dir == os.path.dirname(current_dir):
            break
        else:
            current_dir = os.path.dirname(current_dir)

    return None


def get_app_name_from_yml() -> str:
    file = find_story_yml()
    if file is None:
        return None
    import yaml

    with open(file, 'r') as s:
        return yaml.safe_load(s).pop('app_name')


def get_project_root_dir() -> str:
    return os.path.dirname(find_story_yml())


def get_asyncy_yaml(must_exist=True) -> dict:
    file = find_story_yml()

    # Anybody calling this must've already checked for this file presence.
    if must_exist:
        assert file is not None

    try:
        import yaml

        with open(file, 'r') as s:
            return yaml.safe_load(s)
    except Exception:
        return {}



# -*- coding: utf-8 -*-
import os

from story.commands.logout import logout


def test_logout(runner):
    with runner.runner.isolated_filesystem():
        our_config_file = 'my_super_cool_config_file.json'
        with open(our_config_file, 'w') as f:
            f.write('my_config_file')

        assert os.path.exists(our_config_file)

        result = runner.run(logout, exit_code=0,
                            init_with_config_path=our_config_file,
                            write_default_config=False)

        assert os.path.exists(our_config_file) is False

        assert 'Storyscript CLI installation reset' in result.stdout
        assert 'story login' in result.stdout



# -*- coding: utf-8 -*-
import io
import os
import sys

from setuptools.command.install import install as _install
from setuptools import find_packages, setup, Command

from story.version import version


class Install(_install):
    """Overwrites the default setup.py installation to error on Python < 3.6."""

    def run(self):
        if sys.version_info < (3, 6):
            sys.exit('story requires Python 3.6+')
        _install.run(self)


class PyTest(Command):
    user_options = [("pytest-args=", "a", "Arguments to pass into py.test")]

    def initialize_options(self):
        pass

    def finalize_options(self):
        # Command.finalize_options(self)
        pass

    def run(self):
        import pytest

        errno = pytest.main([])

        sys.exit(errno)


class Format(Command):
    user_options = [('black-args', 'a', 'arguments to pass into black.')]

    def initialize_options(self):
        pass

    def finalize_options(self):
        # Command.finalize_options(self)
        pass

    def run(self):
        os.system('black -S -l 79 story')


classifiers = [
    'Development Status :: 3 - Alpha',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'Intended Audience :: System Administrators',
    'License :: OSI Approved :: MIT License',
    'Natural Language :: English',
    'Programming Language :: Python :: 3.7',
    'Topic :: Office/Business',
    'Topic :: Software Development :: Build Tools',
]

keywords = [
    'storyscript',
    'devops',
    'devtools',
    'microservices',
    'orchestration',
    'serverless',
    'APIs',
    'asyncy',
]

requirements = [
    'click-completion==0.5.1',
    'click-help-colors==0.5',
    'click-spinner==0.1.8',
    'click==7.0',
    'emoji==0.5.0',
    'raven==6.9.0',
    'requests~=2.20',
    'storyscript==0.25.2',
    'websockets~=7.0',
    'texttable~=1.4.0',
    'pyyaml>=4.2',
    'pytz>=2018.5',
    'blindspin',
    'appdirs',
    'xerox',
    'semver',
]

test_requirements = ["pytest"]

# Read the README.md as the long_description.
here = os.path.abspath(os.path.dirname(__file__))
long_description_fname = os.path.join(here, 'README.md')
with io.open(long_description_fname, encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='story',
    version=version,
    description='Storyscript Cloud CLI',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=classifiers,
    download_url='https://github.com/storyscript/cli/archive/'
    + version
    + '.zip',
    keywords=' '.join(keywords),
    author='Storyscript',
    author_email='hello@storyscript.io',
    url='https://docs.storyscript.io/cli',
    license='Apache 2',
    packages=find_packages(exclude=('build.*', 'tests', 'tests.*')),
    include_package_data=True,
    zip_safe=True,
    install_requires=requirements,
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    tests_require=test_requirements,
    extras_require={},
    requires_python='>=3.6.0',
    entry_points={'console_scripts': ['story=story.__main__:cli']},
    cmdclass={'install': Install, 'test': PyTest, 'format': Format},
)



# -*- coding: utf-8 -*-
import os
from unittest import mock

import click

from pytest import mark

from story import api, awesome, cli
from story.helpers import datetime


def test_list(runner, patch, init_sample_app_in_cwd):
    patch.object(datetime, 'parse_psql_date_str', side_effect=[
        'my_app_1_parsed_date',
        'my_app_2_parsed_date'
    ])

    patch.object(datetime, 'reltime', side_effect=[
        'my_app_1_reltime',
        'my_app_2_reltime',
    ])

    ts_app_1 = '2019-06-26T10:31:22.499142+00:00'
    ts_app_2 = '2018-12-17T10:19:56.659736+00:00'

    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()

        from story import api
        from story.commands.apps import apps
        patch.object(api.Apps, 'list',
                     return_value=[
                         {
                             'name': 'my_app_1',
                             'timestamp': ts_app_1,
                             'maintenance': False
                         },
                         {
                             'name': 'my_app_2',
                             'timestamp': ts_app_2,
                             'maintenance': True
                         },
                     ]
                     )

        args = ['list']
        result = runner.run(apps, args=args)

    expected_output = """
   NAME         STATE            CREATED     
============================================
my_app_1   running          my_app_1_reltime
my_app_2   in maintenance   my_app_2_reltime
    """  # noqa (because there's a trailing whitespace in the header)

    assert result.stdout.strip() == expected_output.strip()

    assert datetime.reltime.mock_calls == [
        mock.call('my_app_1_parsed_date'),
        mock.call('my_app_2_parsed_date')
    ]

    assert datetime.parse_psql_date_str.mock_calls == [
        mock.call(ts_app_1),
        mock.call(ts_app_2)
    ]


def test_list_no_apps(runner, patch, init_sample_app_in_cwd):
    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()

        from story import api
        from story.commands.apps import apps
        patch.object(api.Apps, 'list', return_value=[])

        args = ['list']
        result = runner.run(apps, args=args)

    assert 'No application found' in result.stdout
    assert 'story apps create' in result.stdout


def test_do_open(runner, init_sample_app_in_cwd, patch):
    patch.object(click, 'launch')
    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()

        from story.commands.apps import do_open

        result = runner.run(do_open)

    app_url = 'https://my_app.storyscriptapp.com/'
    click.launch.assert_called_with(app_url)
    assert app_url in result.stdout


@mark.parametrize('isatty', [True, False])
def test_url(patch, runner, init_sample_app_in_cwd, isatty):
    patch.object(click, 'launch')

    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()

        from story.commands import apps
        patch.object(apps, '_isatty', return_value=isatty)

        result = runner.run(apps.url)

    app_url = 'https://my_app.storyscriptapp.com/'

    if isatty:
        app_url += '\n'

    assert app_url == result.stdout


@mark.parametrize('all_apps', [True, False])
@mark.parametrize('yes_to_all', [True, False])
@mark.parametrize('app_name', [None, 'my_secret_app'])
def test_destroy(patch, runner, init_sample_app_in_cwd, all_apps,
                 yes_to_all, app_name):
    if all_apps and app_name:  # Invalid combination.
        return

    patch.object(api.Apps, 'destroy')
    if all_apps:
        patch.object(api.Apps, 'list', return_value=[
            {'name': 'my_app_1'},
            {'name': 'my_app_2'},
            {'name': 'my_app_3'},
        ])

    patch.object(cli, 'track')

    args = []

    stdin = 'y\n'

    if all_apps:
        stdin = 'y\ny\ny\n'
        args.append('--all')

    if yes_to_all:
        args.append('--yes')
        stdin = None

    if app_name:
        args.append('-a')
        args.append(app_name)

    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()

        from story.commands import apps
        result = runner.run(apps.destroy, stdin=stdin, args=args)

    if not app_name:
        app_name = 'my_app'  # The default app name, from the current directory

    if all_apps:
        for i in range(1, 4):
            assert f'Destroying application \'my_app_{i}\'' in result.stdout

        assert api.Apps.destroy.mock_calls == [
            mock.call(app='my_app_1'),
            mock.call(app='my_app_2'),
            mock.call(app='my_app_3'),
        ]

        assert cli.track.mock_calls == [
            mock.call('App Destroyed', {'App name': 'my_app_1'}),
            mock.call('App Destroyed', {'App name': 'my_app_2'}),
            mock.call('App Destroyed', {'App name': 'my_app_3'}),
        ]
    else:
        assert f'Destroying application \'{app_name}\'' in result.stdout
        api.Apps.destroy.assert_called_with(app=app_name)
        cli.track.assert_called_with('App Destroyed', {'App name': app_name})


def test_create_inside_an_existing_project(runner, init_sample_app_in_cwd):
    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()
        from story.commands.apps import create
        result = runner.run(create, exit_code=1)

        assert 'There appears to be a Storyscript Cloud project in' \
               f' {os.getcwd()}/story.yml already' in result.stdout


def test_create_with_short_name(runner):
    with runner.runner.isolated_filesystem():
        from story.commands.apps import create
        result = runner.run(create, args=['a'], exit_code=1)

    assert 'The name you specified is too short.' in result.stdout
    assert 'use at least 4 characters' in result.stdout


@mark.parametrize('custom_app_name', [None, 'my_custom_app_name_is_too_cool'])
@mark.parametrize('team', [None, 'my_team_name'])
def test_create(runner, patch, custom_app_name, team):
    args = []

    if custom_app_name:
        app_name = custom_app_name
        args.append(app_name)
    else:
        app_name = 'my_app_name_different_than_default'
        patch.object(awesome, 'new', return_value=app_name)

    if team:
        args.append('--team')
        args.append(team)

    from story.commands import apps
    patch.object(apps, 'create_story_yaml', side_effect=apps.create_story_yaml)
    patch.object(api.Apps, 'create')
    patch.object(cli, 'track')

    with runner.runner.isolated_filesystem():
        result = runner.run(apps.create, args=args, exit_code=0)

        with open('story.yml') as f:
            actual_contents_of_story_yml = f.read()

    assert 'Creating story.yml…' in result.stdout
    assert f'App Name: {app_name}' in result.stdout
    assert f'App URL: https://{app_name}.storyscriptapp.com/' in result.stdout
    assert 'We hope you enjoy your deployment experience' in result.stdout

    assert actual_contents_of_story_yml == f'app_name: {app_name}\n'

    apps.create_story_yaml.assert_called_with(app_name)
    api.Apps.create.assert_called_with(name=app_name, team=team)
    cli.track.assert_called_with('App Created', {'App name': app_name})


def test_init_inside_an_existing_project(runner, init_sample_app_in_cwd):
    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()
        from story.commands import apps
        result = runner.run(apps.init, args=['myapp'], exit_code=1)

        assert 'There appears to be a Storyscript Cloud project in' \
               f' {os.getcwd()}/story.yml already' in result.stdout


def test_init(runner, patch):
    app_name = 'myapp'
    patch.object(api.Apps, 'get_uuid_from_hostname')
    from story.commands import apps
    patch.object(apps, 'create_story_yaml', side_effect=apps.create_story_yaml)
    with runner.runner.isolated_filesystem():
        result = runner.run(apps.init, args=[app_name], exit_code=0)
        with open('story.yml') as f:
            actual_contents_of_story_yml = f.read()

    apps.create_story_yaml.assert_called_with(app_name)
    api.Apps.get_uuid_from_hostname.assert_called_with(app_name)

    assert 'Creating story.yml…' in result.stdout
    assert f'App Name: {app_name}' in result.stdout
    assert f'App URL: https://{app_name}.storyscriptapp.com/' in result.stdout
    assert 'We hope you enjoy your deployment experience' in result.stdout

    assert actual_contents_of_story_yml == f'app_name: {app_name}\n'



# -*- coding: utf-8 -*-
from .. import cli


@cli.cli.command()
def logout():
    """Logout from the Storyscript Cloud.
    This command will only remove your login credentials, and not your apps."""
    cli.reset()
